# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, List, Any, Mapping
import numpy
from cellpose import models
from skimage.measure import regionprops
from dask.distributed import get_worker
import torch
from scip.utils.util import copy_without
from distributed import get_client


def segment_block(
    events: List[Mapping[str, Any]],
    *,
    channel_indices: List[int] = None,
    parent_channel_index: int,
    dapi_channel_index: int,
    gpu_accelerated: Optional[bool] = False,
    cell_diameter: Optional[int] = None,
    flow_threshold: Optional[float] = 0.4,
    **kwargs
) -> List[dict]:

    if len(events) == 0:
        return events

    w = get_worker()
    if hasattr(w, "cellpose"):
        model = w.cellpose
    else:
        if gpu_accelerated:
            gpu_id = list(get_client().scheduler_info()["workers"].keys()).index(w.address)
            device = torch.device(f'cuda:{gpu_id}')
            model = models.Cellpose(gpu=True, device=device, model_type='cyto2')
        else:
            model = models.Cellpose(gpu=False, model_type='cyto2')
        w.cellpose = model

    parents, _, _, _ = model.eval(
        x=[e["pixels"][[parent_channel_index, dapi_channel_index]] for e in events],
        channels=[1, 2],
        diameter=cell_diameter,
        batch_size=128,
        flow_threshold=flow_threshold
    )

    if channel_indices is None:
        channel_indices = range(len(events[0]["pixels"]))

    children = []
    for channel_index in channel_indices:
        if channel_index == parent_channel_index:
            continue

        o, _, _, _ = model.eval(
            x=[e["pixels"][[channel_index, dapi_channel_index]] for e in events],
            channels=[1, 2],
            diameter=cell_diameter,
            batch_size=128,
            flow_threshold=flow_threshold
        )
        children.append((channel_index, o))

    for e_i, event in enumerate(events):

        labeled_mask = numpy.repeat(parents[e_i][numpy.newaxis], event["pixels"].shape[0], axis=0)

        for channel_index, child in children:

            # assign over-segmented children to parent objects
            mask = numpy.zeros_like(parents[e_i])
            for i in numpy.unique(parents[e_i])[1:]:
                idx, counts = numpy.unique(
                    child[e_i][parents[e_i] == i], return_counts=True)
                idx, counts = idx[1:], counts[1:]  # skip zero (= background)

                idx = idx[(counts / (parents[e_i] == i).sum()) > 0.1]
                mask[numpy.isin(child[e_i], idx) & (parents[e_i] == i)] = i

            labeled_mask[channel_index] = mask

        event["mask"] = labeled_mask

    return events


def to_events(
    events: List[Mapping[str, Any]],
    *,
    group_keys: Optional[List[str]] = None,
    parent_channel_index: int,
    **kwargs
):
    """Converts the segmented objects into a list of dictionaries that can be converted
    to a dask.bag.Bag. The dictionaries contain the pixel information for one detected cell,
    or event, and the meta data of that event.
    """

    newevents = []
    for event in events:

        labeled_mask = event["mask"]
        cells = labeled_mask[parent_channel_index]
        cell_regions = regionprops(cells)

        if group_keys is not None:
            group = "_".join([str(event[k]) for k in group_keys])
        else:
            group = None

        for props in cell_regions:

            bbox = props.bbox

            mask = labeled_mask[:, bbox[0]:bbox[2], bbox[1]:bbox[3]] == props.label
            combined_mask = cells[bbox[0]:bbox[2], bbox[1]:bbox[3]] == props.label

            regions = []
            for m in mask:
                regions.append(int(m.any()))

            newevent = copy_without(event=event, without=["mask", "pixels"])
            newevent["pixels"] = event["pixels"][:, bbox[0]: bbox[2], bbox[1]:bbox[3]]
            newevent["combined_mask"] = combined_mask
            newevent["mask"] = mask
            newevent["group"] = group
            newevent["bbox"] = tuple(bbox)
            newevent["regions"] = regions
            newevent["background"] = numpy.zeros(
                shape=(event["pixels"].shape[0],), dtype=float)
            newevent["combined_background"] = numpy.zeros(
                shape=(event["pixels"].shape[0],), dtype=float)
            newevent["id"] = props.label

            newevents.append(newevent)

    return newevents



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from skimage.morphology import (
    remove_small_holes, remove_small_objects, label)
from skimage.filters import threshold_li, sobel, gaussian
from scipy.stats import normaltest
from scip.utils.util import check, copy_without
from scip.masking import mask_predicate


@check
def get_mask(el, main_channel, smooth):

    regions = [0] * len(el["pixels"])
    mask = numpy.full(shape=el["pixels"].shape, dtype=bool, fill_value=False)

    # load over channels, starting with main_channel
    arr = numpy.arange(len(el["pixels"]))
    for dim in [main_channel] + numpy.delete(arr, arr == main_channel).tolist():
        cc = 0

        x = el["pixels"][dim]
        if (normaltest(x.ravel()).pvalue < 0.05):
            x = gaussian(x, sigma=smooth[dim])
            x = sobel(x)
            x = gaussian(x, sigma=smooth[dim] * 2)
            x = threshold_li(x) < x
            x = remove_small_holes(x, area_threshold=(x.shape[0] * x.shape[1]) / 4)
            x = remove_small_objects(x, min_size=20)
            x = label(x)

            mask[dim], cc = x > 0, x.max()
        elif dim == main_channel:
            out = copy_without(el, without=["pixels"])
            out["regions"] = regions
            out["mask"] = mask
            return out

        regions[dim] = cc

    out = el.copy()
    out["mask"] = mask
    out["regions"] = regions

    return out


def create_masks_on_bag(bag, main_channel, smooth):

    def threshold_masking(partition):
        return [
            mask_predicate(get_mask(p, main_channel, smooth), main_channel)
            for p in partition
        ]

    bag = bag.map_partitions(threshold_masking)
    return bag



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

import dask
from PIL import Image
import dask.bag
import dask.dataframe
import numpy
from pathlib import Path
import pandas


def load_image(event, channels=None):
    """
    Load an image from a certain path

    Args:
        path (str): path of image
        channels (list, optional): image channels to load. Defaults to None.

    Returns:
        dict: dictionary containing pixel values (ndarray) and path for each image
    """
    im = Image.open(event["path"])

    if channels is None:
        channels = range(im.n_frames)

    arr = numpy.empty(shape=(len(channels), im.height, im.width), dtype=float)
    for i, c in enumerate(channels):
        im.seek(c)
        arr[i] = numpy.array(im)

    newevent = event.copy()
    newevent["pixels"] = arr
    return newevent


def bag_from_directory(path, channels, partition_size):
    """
    Construct delayed ops for all tiffs in a directory

    Args:
        path (str): Directory to find tiffs

    Returns:
        dask.bag: bag containing dictionaries with image data
    """

    def load_image_partition(partition):
        return [load_image(event, channels) for event in partition]

    events = []
    for i, p in enumerate(Path(path).glob("**/*.tiff")):
        events.append(dict(path=str(p), group=str(p.parent)))

    meta = pandas.DataFrame.from_records(data=events)
    meta.columns = [f"meta_{c}" for c in meta.columns]
    meta = dask.dataframe.from_pandas(meta, chunksize=partition_size)

    bag = dask.bag.from_sequence(events, partition_size=partition_size)
    bag = bag.map_partitions(load_image_partition)

    return bag, meta



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

import socket
from dask.distributed import Client
from pathlib import Path
import yaml
from pkg_resources import resource_stream
import logging
import math
import shutil
import click
from datetime import datetime, timedelta
import dask
from scip._version import get_versions

MODES = ["local", "jobqueue", "mpi", "external"]


class ClientClusterContext:

    def __init__(
            self,
            *,
            mode="local",
            n_workers=12,
            n_nodes=1,
            port=8787,
            local_directory=None,
            memory=None,
            cores=None,
            job_extra=[],
            walltime="01:00:00",
            threads_per_process=None,
            project=None,
            gpu=0,
            scheduler_adress: str = None,
    ):
        self.mode = mode
        self.n_workers = n_workers
        self.port = port
        self.n_nodes = n_nodes
        self.local_directory = local_directory
        self.memory = memory
        self.cores = cores
        self.job_extra = job_extra
        self.walltime = walltime
        self.threads_per_process = threads_per_process
        self.project = project
        self.gpu = gpu
        self.scheduler_adress = scheduler_adress

    def __enter__(self):
        if self.mode == "local":
            from dask.distributed import LocalCluster
            with dask.config.set({"distributed.worker.resources.cellpose": 1}):
                self.cluster = LocalCluster(
                    n_workers=self.n_workers, threads_per_worker=self.threads_per_process,
                    processes=True
                )
            self.client = Client(self.cluster)
        elif self.mode == "jobqueue":
            from dask_jobqueue import PBSCluster

            assert (Path.home() / "logs").exists(), "Make sure directory\
                 'logs' exists in your home dir"

            t = datetime.strptime(self.walltime, "%H:%M:%S")
            seconds = timedelta(hours=t.hour, minutes=t.minute, seconds=t.second).total_seconds()
            extra = ["--lifetime", f"{seconds}s"]
            if self.threads_per_process is not None:
                extra = ["--nthreads", self.threads_per_process]

            mb_needed = math.ceil(self.memory * 1073.74)
            self.cluster = PBSCluster(
                cores=self.cores,
                memory=f"{self.memory}GiB",
                processes=self.n_workers,
                resource_spec=f"nodes=1:ppn={self.cores},mem={mb_needed}mb",
                project=self.project,
                local_directory=self.local_directory,
                walltime=self.walltime,
                extra=extra,
                job_extra=self.job_extra,
                scheduler_options={
                    'dashboard_address': None if self.port is None else f':{self.port}'},
                death_timeout=10 * 60
            )

            self.cluster.scale(jobs=self.n_nodes)
            self.client = Client(self.cluster)
        elif self.mode == "mpi":
            from dask_mpi import initialize
            from mpi4py import MPI

            worker_options = {}
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()

            if (self.gpu > 0) and rank in range(2, 2 + self.gpu):
                worker_options["resources"] = {'cellpose': 1}

            is_client = initialize(
                dashboard=True,
                dashboard_address=None if self.port is None else f':{self.port}',
                nthreads=self.threads_per_process,
                local_directory=self.local_directory,
                memory_limit=int(self.memory * 1e9),
                worker_class='distributed.Nanny',
                worker_options=worker_options,
                exit=False
            )

            if is_client:
                self.client = Client()
                self.client.wait_for_workers(n_workers=self.n_workers)
        elif self.mode == "external":
            assert self.scheduler_adress is not None, "Adress must be set in external mode."
            self.client = Client(address=self.scheduler_adress)

        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.mode == "mpi" and hasattr(self, "client"):
            from dask_mpi import send_close_signal
            send_close_signal()
        elif hasattr(self, "client"):
            self.client.close()

        if exc_type is not None:
            logging.getLogger(__name__).error(
                "Exception in context: %s, %s", exc_type, str(exc_value))

    def wait(self):
        if self.mode in ["local", "mpi"]:
            n_workers = self.n_workers
        elif self.mode == "jobqueue":
            n_workers = self.n_workers * self.n_nodes

        self.client.wait_for_workers(n_workers=n_workers)


def load_yaml_config(path):
    with open(path) as fh:
        return yaml.load(fh, Loader=yaml.FullLoader)


def configure_logging(output, debug):
    with resource_stream(__name__, 'logging.yml') as stream:
        loggingConfig = yaml.load(stream, Loader=yaml.FullLoader)

    for k, v in loggingConfig["handlers"].items():
        if k == "file":
            v["filename"] = str(output / v["filename"])
        if (k == "console") and debug:
            v["level"] = "DEBUG"

    logging.config.dictConfig(loggingConfig)


def make_output_dir(output, headless):
    should_remove = True

    if (not headless) and output.exists():
        should_remove = click.prompt(
            f"{str(output)} exists. Overwrite contents? [Y/n]",
            type=str, show_default=False, default="Y"
        ) == "Y"

        if not should_remove:
            raise FileExistsError(f"{str(output)} exists and should not be removed. Exiting.")
    if should_remove and output.exists():
        shutil.rmtree(output)
    if not output.exists():
        output.mkdir(parents=True)


def copy_without(event, without=[]):
    return {
        k: v for k, v in event.items()
        if k not in without
    }


def check(func):
    def inner(sample, *args, **kwargs):
        if "pixels" in sample:
            return func(sample, *args, **kwargs)
        else:
            return sample
    return inner


def prerun(context, paths, output, headless, debug, mode, gpu, n_partitions, n_threads):

    make_output_dir(output, headless=headless)

    configure_logging(output, debug)
    logger = logging.getLogger("scip")

    version = get_versions()["version"]
    logger.info(f"SCIP version {version}")

    t = datetime.utcnow().isoformat(timespec="seconds")
    logger.info(f"Starting at {t}")
    logger.info(f"Running pipeline for {','.join(paths)}")

    n_workers = len(context.client.scheduler_info()["workers"])
    logger.info(f"Running with {n_workers} workers and {n_threads} threads per worker")
    logger.info(f"Mode: {mode}")
    logger.info(f"GPUs: {gpu}")
    logger.info(f"Number of partitions: {n_partitions}")
    logger.info(f"Output is saved in {str(output)}")

    host = context.client.run_on_scheduler(socket.gethostname)
    port = context.client.scheduler_info()['services']['dashboard']
    logger.info(f"Dashboard -> ssh -N -L {port}:{host}:{port}")



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Optional, List

import time
import os
import logging
import logging.config
from pathlib import Path
from importlib import import_module

import click
import dask.bag
import dask.dataframe
import dask.dataframe.multi
import pandas

from scip.loading import load_meta, load_pixels
from scip.utils.util import copy_without, prerun
from scip.utils import util  # noqa: E402
from scip.features import compute_features  # noqa: E402
from scip.masking import mask
from scip.segmentation import segment
from scip.projection import project_block_partition
from scip._version import get_versions

# dask issues a warning during normalization
# when initializing the map-reduce operation
# this warning can only be fixed by fixing dask
import warnings
warnings.simplefilter("ignore", category=FutureWarning)


@dask.delayed
def channel_boundaries(quantiles, *, config, output):
    data = []
    index = []
    for k, v in quantiles:
        index.append(k)
        out = {}
        for channel, r in zip(config["load"]["channel_names"], v):
            out[f"{channel}_min"] = r[0]
            out[f"{channel}_max"] = r[1]
        data.append(out)
    pandas.DataFrame(data=data, index=index).to_csv(str(output / "channel_boundaries.csv"))


def main(  # noqa: C901
    *,
    paths: List[str],
    output: Path,
    config: str,
    mode: str,
    limit: Optional[int] = -1,
    with_replacement: Optional[bool] = False,
    n_partitions: Optional[int] = 10,
    n_workers: Optional[int] = 1,
    n_nodes: Optional[int] = 1,
    n_cores: Optional[int] = None,
    n_threads: Optional[int] = 1,
    memory: Optional[int] = 1,
    walltime: Optional[str] = None,
    job_extra: Optional[str] = "",
    local_directory: Optional[str] = "tmp",
    headless: Optional[bool] = False,
    port: Optional[int] = 8787,
    debug: Optional[bool] = False,
    timing: Optional[Any] = None,
    gpu: Optional[int] = 0,
    scheduler_adress: Optional[str] = None
):
    with util.ClientClusterContext(
            n_workers=n_workers,
            mode=mode,
            port=port,
            n_nodes=n_nodes,
            local_directory=local_directory,
            cores=n_cores,
            memory=memory,
            walltime=walltime,
            job_extra=job_extra,
            threads_per_process=n_threads,
            gpu=gpu,
            scheduler_adress=scheduler_adress
    ) as context:

        if not hasattr(context, "client"):
            return

        output = Path(output)
        prerun(context, paths, output, headless, debug, mode, gpu, n_partitions, n_threads)

        logger = logging.getLogger("scip")

        # if timing is set, wait for the cluster to be fully ready
        # to isolate cluster startup time from pipeline execution
        if timing is not None:
            logger.debug("waiting for all workers")
            context.wait()

        logger.debug("timer started")
        start = time.time()

        config = util.load_yaml_config(config)
        assert all([
            k in config
            for k in [
                "filter",
                "normalization",
                "load",
                "mask",
                "segment",
                "project",
                "feature_extraction",
                "export",
                "illumination_correction"
            ]]), "Config is incomplete."
        logger.info(f"Running with following config: {config}")

        assert "channels" in config["load"], "Please specify what channels to load"
        channels = config["load"]["channels"]
        channel_names = config["load"]["channel_names"]
        assert len(channels) == len(channel_names), "Please specify a name for each channel"

        loader_module = import_module('scip.loading.%s' % config["load"]["format"])
        # with dask.config.set(**{'array.slicing.split_large_chunks': False}):

        meta = load_meta(
            paths=paths,
            kwargs=config["load"]["kwargs"] or dict(),
            loader_module=loader_module
        ).persist()
        images = meta.repartition(npartitions=n_partitions)

        images = load_pixels(
            bag=images,
            channels=channels,
            kwargs=config["load"]["kwargs"] or dict(),
            loader_module=loader_module
        )

        if config["project"] is not None:
            project_method = config["project"]["method"]
            project_kw = config["project"]["settings"]
            project_block = import_module('scip.projection.%s' % project_method).project_block
            images = images.map_partitions(
                project_block_partition, proj=project_block, **project_kw)

        if config["illumination_correction"] is not None:
            method = config["illumination_correction"]["method"]
            key = config["illumination_correction"]["key"]
            correct = import_module('scip.illumination_correction.%s' % method).correct

            ill_corr_output = None
            if config["illumination_correction"]["export"]:
                ill_corr_output = output
            images = correct(
                images=images,
                key=key,
                ngroups=len(meta.distinct(key).compute()),
                output=ill_corr_output,
                **config["illumination_correction"]["settings"],
            )

        if config["segment"] is not None:
            images = segment(
                images=images,
                method=config["segment"]["method"],
                settings=config["segment"]["settings"],
                export=config["segment"]["export"],
                output=output,
                gpu=gpu,
                loader_module=loader_module
            )

        if limit > 0:
            from dask.bag.random import sample, choices
            if with_replacement:
                images = choices(images, k=limit)
            else:
                images = sample(images, k=limit)

        if config["mask"] is not None:
            images_dict = mask(
                images=images,
                methods=config["mask"]["methods"],
                combined_indices=config["mask"]["combined_indices"]
            )
        else:
            images_dict = dict(no=images)

        loader_meta = loader_module.get_loader_meta(
            **(config["load"]["kwargs"] or dict()))
        dataframes = []
        futures = []
        for prefix, images in images_dict.items():
            if config["filter"] is not None:
                filter_module = import_module('scip.filter.%s' % config["filter"]["name"])

                images = images.map_partitions(filter_module.feature_partition)
                images = images.map(copy_without, without=["pixels"]).persist()
                filter_items = filter_module.item(images)

                images = images.map(filter_module.predicate, **filter_items)

                images = images.map_partitions(
                    loader_module.reload_image_partition,
                    channels=channels,
                    **(config["load"]["kwargs"] or dict())
                )

            quantiles = None
            if config["normalization"] is not None:
                logger.debug("performing normalization")
                from scip.normalization import quantile_normalization  # noqa: E402
                images, quantiles = quantile_normalization.quantile_normalization(
                    images,
                    config["normalization"]["lower"],
                    config["normalization"]["upper"],
                    len(channels)
                )
                futures.append(channel_boundaries(quantiles, config=config, output=output))

            logger.debug("computing features")

            pref = None if prefix == "no" else prefix
            types = config["feature_extraction"]
            if prefix != "no":
                types = types[prefix]

            bag_df = compute_features(
                images=images,
                channel_names=channel_names,
                loader_meta=loader_meta,
                prefix=pref,
                types=types
            )

            dataframes.append(bag_df)

            # change loader_meta dict so that meta keys are only added once
            # in case multiple features sets are computed.
            # Leave in the regions keys as they might change between masks
            loader_meta = {k: v for k, v in loader_meta.items() if "regions" in k}

        # partitions never change between masks so we can ignore unknown divisions
        bag_df = dask.dataframe.multi.concat(dataframes, axis=1, ignore_unknown_divisions=True)
        bag_df = bag_df.repartition(npartitions=10)

        filename = config["export"]["filename"]
        export_module = import_module('scip.export.%s' % config["export"]["format"])
        futures.append(export_module.export(df=bag_df, filename=filename, output=output))

        dask.compute(*futures, traverse=False, optimize_graph=False)

        if debug:
            context.client.profile(filename=str(output / "profile.html"))

    runtime = time.time() - start
    logger.info(f"Full runtime {runtime:.2f}")
    return runtime


def _print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(get_versions()['version'])
    ctx.exit()


@click.command(
    name="Scalable imaging pipeline",
    context_settings=dict(show_default=True)
)
@click.option("--port", "-d", type=int, default=None, help="dask dashboard port")
@click.option("--debug", envvar="DEBUG", is_flag=True, help="sets logging level to debug")
@click.option(
    "--mode", default="local", type=click.Choice(util.MODES),
    help="In which mode to run Dask")
@click.option(
    "--n-workers", "-j", type=int, default=-1,
    help="Number of workers in the LocalCluster or per node")
@click.option(
    "--n-nodes", "-n", type=int, default=1,
    help="Number of nodes started")
@click.option(
    "--n-cores", "-c", type=click.IntRange(min=1), default=1,
    help="Number of cores available per node in the cluster")
@click.option(
    "--n-threads", "-t", type=click.IntRange(min=1), default=None,
    help="Number of threads per worker process")
@click.option(
    "--memory", "-m", type=click.IntRange(min=1), default=4,
    help="Amount of memory available per node in the cluster")
@click.option(
    "--limit", "-i", type=click.IntRange(min=-1), default=-1,
    help="Amount of images to sample randomly from the dataset. -1 means no sampling.")
@click.option(
    "--with-replacement", type=bool, is_flag=True, default=False,
    help="Enable sampling with replacement. Has no effect is limit is set to default.")
@click.option(
    "--walltime", "-w", type=str, default="01:00:00",
    help="Expected required walltime for the job to finish")
@click.option(
    "--job-extra", "-e", type=str, multiple=True, default=[],
    help="Extra arguments for job submission")
@click.option(
    "--headless", default=False, is_flag=True,
    help="If set, the program will never ask for user input")
@click.option(
    "--n-partitions", "-s", default=10, type=click.IntRange(min=1),
    help="Set number of partitions")
@click.option(
    "--scheduler-adress", default=None, type=str,
    help="Adress of scheduler to connect to."
)
@click.option("--timing", default=None, type=click.Path(dir_okay=False))
@click.option(
    "--gpu", default=0, type=click.IntRange(min=0), help="Specify the amount of available GPUs")
@click.option(
    "--local-directory", "-l", default=None, type=click.Path(file_okay=False, exists=True))
@click.option(
    "-V", "--version", default=False, is_flag=True, is_eager=True,
    expose_value=False, callback=_print_version,
    help="Display version information"
)
@click.argument("output", type=click.Path(file_okay=False))
@click.argument("config", type=click.Path(dir_okay=False, exists=True))
@click.argument("paths", nargs=-1, type=str)
def cli(**kwargs):
    """Intro documentation
    """

    # noop if no paths are provided
    if len(kwargs["paths"]) == 0:
        return

    def check(p: str) -> bool:
        if p.startswith("hdfs"):
            return True
        return os.path.isabs(p)
    if kwargs["mode"] == "external":
        assert check(kwargs["output"]), "Output path must be absolute in external mode."
        err = "Paths must be absolute in external mode."
        assert all([check(p) for p in kwargs["paths"]]), err

    runtime = main(**kwargs)

    if runtime is not None:
        timing = kwargs["timing"]
        if timing is not None:
            import json
            with open(timing, "w") as fp:
                json.dump({**kwargs, **dict(runtime=runtime)}, fp)
            logging.getLogger("scip").info(f"Timing output written to {timing}")


if __name__ == "__main__":
    cli()



import numpy
from scip.masking import mask_predicate


def _create_circular_mask(i):

    c, h, w = i.shape

    center = (int(w / 2), int(h / 2))
    radius = min(w, h) // 3

    Y, X = numpy.ogrid[:h, :w]
    dist_from_center = numpy.sqrt((X - center[0])**2 + (Y - center[1])**2)

    mask = dist_from_center <= radius
    return numpy.repeat(mask[numpy.newaxis], repeats=c, axis=0)


def get_mask(el):

    regions = [1] * len(el["pixels"])
    mask = _create_circular_mask(el["pixels"])

    newel = el.copy()
    newel["mask"] = mask
    newel["regions"] = regions
    return newel


def create_masks_on_bag(bag, main_channel):

    def circle_masking(partition):
        return [
            mask_predicate(get_mask(p), main_channel)
            for p in partition
        ]

    bag = bag.map_partitions(circle_masking)
    return bag



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

import numpy
from scipy.stats import normaltest
from scip.utils.util import check, copy_without
from scip.masking import mask_predicate
from skimage.morphology import white_tophat, disk, label, binary_dilation
from skimage.filters import threshold_minimum


@check
def get_mask(el, main_channel, spotsize):

    regions = [0] * len(el["pixels"])
    mask = numpy.full(shape=el["pixels"].shape, dtype=bool, fill_value=False)

    # load over channels, starting with main_channel
    arr = numpy.arange(len(el["pixels"]))
    for dim in [main_channel] + numpy.delete(arr, arr == main_channel).tolist():
        cc = 0

        x = el["pixels"][dim]
        if (normaltest(x.ravel()).pvalue < 0.05):

            x = el["pixels"][dim]
            x = white_tophat(x, footprint=disk(spotsize))

            for nbins in [256, 512, 1024]:
                try:
                    x = threshold_minimum(x, nbins=nbins) < x
                    x = binary_dilation(x, footprint=disk(2))
                    x = label(x)
                    mask[dim], cc = x > 0, x.max()
                    break
                except RuntimeError:
                    pass
        elif dim == main_channel:
            out = copy_without(el, without=["pixels"])
            out["regions"] = regions
            out["mask"] = mask
            return out

        regions[dim] = cc

    out = el.copy()
    out["mask"] = mask
    out["regions"] = regions

    return out


def create_masks_on_bag(bag, main_channel, spotsize):

    def spot_masking(partition):
        return [
            mask_predicate(get_mask(p, main_channel, spotsize), main_channel)
            for p in partition
        ]

    bag = bag.map_partitions(spot_masking)
    return bag



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from skimage.segmentation import watershed
from skimage.filters import sobel, threshold_otsu
from skimage import morphology
from skimage.restoration import denoise_nl_means
import numpy
from scip.masking import mask_post_process


def get_mask(el, noisy_channels):

    image = el["pixels"]
    mask = numpy.empty(shape=image.shape, dtype=bool)
    regions = []

    for dim in range(len(image)):

        if dim in noisy_channels:
            image[dim] = denoise_nl_means(image[dim], patch_size=2, patch_distance=1)

        elev_map = sobel(image[dim])
        closed = morphology.closing(elev_map, footprint=morphology.disk(2))

        # markers = numpy.zeros_like(image[dim])
        # markers[closed < numpy.quantile(closed, 0.7)] = 1
        # markers[closed > numpy.quantile(closed, 0.95)] = 2
        markers = numpy.zeros_like(image[dim])
        thresh = threshold_otsu(closed)
        markers[closed < thresh - thresh * 0.5] = 1
        markers[closed > thresh + thresh * 0.5] = 2

        segmentation = watershed(image[dim], markers, compactness=1)

        if segmentation.max() == 0:
            mask[dim] = False
            regions.append(0)
        else:
            mask[dim], cc = mask_post_process(segmentation == segmentation.max())
            regions.append(cc)

    out = el.copy()
    out["mask"] = mask
    out["regions"] = regions

    return out


def create_masks_on_bag(bag, noisy_channels, **kwargs):

    def watershed_masking(partition):
        return [get_mask(p, noisy_channels) for p in partition]

    bag = bag.map_partitions(watershed_masking)

    return bag



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

"""
Data loaders for various image formats. All data loaders take a path as input and produce a
lazy collection containing all fields of view with meta data.
"""

from typing import List, Mapping, Any, Tuple

import dask.bag
from pathlib import Path


def load_meta(
    *,
    paths: List[str],
    kwargs: Mapping[str, Any] = {},
    loader_module
) -> Tuple[dask.bag.Bag, Mapping[str, type]]:

    bags = []
    for path in paths:
        assert Path(path).exists(), f"{path} does not exist."

        meta = loader_module.meta_from_directory(path=path, **kwargs)
        bag = dask.bag.from_delayed(meta)
        bags.append(bag)

    return dask.bag.concat(bags)


def load_pixels(
    *,
    bag: dask.bag.Bag,
    channels: List[int],
    kwargs: Mapping[str, Any] = {},
    loader_module
) -> dask.bag.Bag:

    bag = loader_module.load_pixels(bag, channels=channels, **kwargs)

    return bag



import dask.dataframe
from pathlib import Path
from concurrent.futures import Future


def export(
    *,
    df: dask.dataframe.DataFrame,
    output: Path,
    filename: str
) -> Future:
    """Exports dataframe to one AnnData .parquet-file per partition.

    Keyword args:
        df: Dataframe to be exported.
        output: Path to directory where objects should be stored.
        filename: Filename to give to partition
          objects (will be named with format string {filename}.{partition}.h5ad).

    Returns:
        Future that represents the export task.
    """

    return df.to_parquet(
        str(output),
        name_function=lambda x: f"{filename}.{x}.parquet",
        write_metadata_file=False,
        engine="pyarrow"
    )



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from skimage.filters import sobel
from skimage import morphology
import numpy
from scip.masking import mask_post_process


def get_mask(el):

    image = el["pixels"]
    mask = numpy.empty(shape=image.shape, dtype=bool)

    for dim in range(len(image)):

        elev_map = sobel(image[dim])
        closed = morphology.closing(elev_map, footprint=morphology.disk(2))

        segmentation = numpy.full(shape=closed.shape, fill_value=False, dtype=bool)
        segmentation[closed > numpy.quantile(closed, 0.9)] = True

        if segmentation.max() == 0:
            mask[dim] = False
        else:
            segmentation = segmentation == segmentation.max()
            mask[dim] = mask_post_process(segmentation)

    out = el.copy()
    out["mask"] = mask

    return out


def create_masks_on_bag(bag, **kwargs):

    def watershed_masking(partition):
        return [get_mask(p) for p in partition]

    bag = bag.map_partitions(watershed_masking)

    return dict(watershed=bag)



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Mapping, Any, List

import numpy

import dask.bag

from skimage.measure import regionprops
from skimage.morphology import remove_small_objects, label, remove_small_holes
from skimage.segmentation import expand_labels

from scip.utils.util import copy_without, check
from numba import njit
from importlib import import_module


def mask(
    *,
    images: dask.bag.Bag,
    methods: Mapping[str, Any],
    combined_indices: List[int]
) -> Mapping[str, dask.bag.Bag]:
    images_dict = {}

    for method in methods:
        masking_module = import_module('scip.masking.%s' % method["method"])

        tmp_images = masking_module.create_masks_on_bag(
            images,
            main_channel=method["bbox_channel_index"],
            **(method["kwargs"] or dict())
        )

        tmp_images = tmp_images.map_partitions(
            remove_regions_touching_border_partition,
            bbox_channel_index=method["bbox_channel_index"]
        )

        tmp_images = tmp_images.map_partitions(bounding_box_partition)

        # mask is applied and background values are computed
        tmp_images = tmp_images.map_partitions(
            apply_mask_partition,
            combined_indices=combined_indices
        )

        images_dict[method["name"]] = tmp_images

    return images_dict


@njit(cache=True)
def _touching_border(mask):

    limit = 10

    if mask[0, :].sum() > limit:
        return True
    if mask[-1, :].sum() > limit:
        return True
    if mask[:, 0].sum() > limit:
        return True
    if mask[:, -1].sum() > limit:
        return True

    return False


def mask_predicate(s, bbox_channel_index):

    # a mask should be present in the bbox_channel_index
    if not numpy.any(s["mask"][bbox_channel_index]):
        return copy_without(s, without=["mask", "pixels"])

    # only one connected component should be found in the bbox channel
    if s["regions"][bbox_channel_index] != 1:
        return copy_without(s, without=["mask", "pixels"])

    # mask should not touch the border of the image
    if _touching_border(s["mask"][bbox_channel_index]):
        return copy_without(s, without=["mask", "pixels"])

    return s


def _regions_touching(arr):

    limit = 10

    # get all unique indices in the arr edges

    top = arr[0, :]
    bottom = arr[-1, :]
    left = arr[:, 0]
    right = arr[:, -1]
    a = numpy.concatenate((top, bottom, left, right))
    idx, counts = numpy.unique(a, return_counts=True)

    if idx[0] == 0:
        return idx[1:][counts[1:] > limit]
    else:
        return idx[counts > limit]


@check
def remove_regions_touching_border(p, bbox_channel_index):
    mask = numpy.empty_like(p["mask"])
    for i in range(len(mask)):
        if i == bbox_channel_index:
            mask[i] = p["mask"][i]
            continue

        x = label(p["mask"][i])
        indices = _regions_touching(x)
        mask[i] = p["mask"][i] * ~numpy.isin(x, indices)

    newevent = copy_without(p, without=["mask"])
    newevent["mask"] = mask

    return newevent


def remove_regions_touching_border_partition(part, bbox_channel_index):
    return [remove_regions_touching_border(p, bbox_channel_index) for p in part]


def apply_mask_partition(part, combined_indices=None):
    return [apply(p, combined_indices) for p in part]


@check
def apply(sample: Mapping[str, Any], combined_indices: List[int] = None):
    """
    Apply binary mask on every channel

    Args:
        sample: dictionary containg image data
        combined_indices: list of indices to be included in combined mask

    Returns:
        dict: dictionary including applied mask
    """

    img = sample["pixels"]
    mask = sample["mask"]

    i = numpy.s_[:] if combined_indices is None else combined_indices
    combined_mask = numpy.sum(mask[i], axis=0) > 0
    background = numpy.empty(shape=(len(img),), dtype=float)
    combined_background = numpy.empty(shape=(len(img),), dtype=float)

    # Multiply image with mask to set background to zero
    for i in range(img.shape[0]):
        if numpy.any(~mask[i]):
            background[i] = img[i][~mask[i]].mean()
        else:
            background[i] = 0
        combined_background[i] = img[i][~combined_mask].mean()

    output = sample.copy()
    output["combined_mask"] = combined_mask
    output["background"] = background.tolist()
    output["combined_background"] = combined_background.tolist()

    return output


def bounding_box_partition(part):
    return [get_bounding_box(event) for event in part]


@check
def get_bounding_box(event):
    minr, minc, maxr, maxc = event["pixels"].shape[1], event["pixels"].shape[2], 0, 0
    for mask in event["mask"]:
        if numpy.any(mask):
            b = regionprops(mask.astype(int))[0].bbox
            minr = min(b[0], minr)
            minc = min(b[1], minc)
            maxr = max(b[2], maxr)
            maxc = max(b[3], maxc)

    newevent = event.copy()
    newevent["bbox"] = [minr, minc, maxr, maxc]

    return newevent


def mask_post_process(mask):
    mask = remove_small_holes(mask, area_threshold=300)
    mask = expand_labels(label(mask), distance=1)
    mask = remove_small_objects(mask > 0, min_size=20)
    mask = label(mask)

    return mask > 0, mask.max()



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import dask
import dask.bag
import dask.array
from scip.utils.util import check, copy_without


def get_distributed_minmax(bag, nchannels):  # noqa: C901

    def combine_extent_partition(a, b):

        if "pixels" not in b:
            return a

        if "mask" in b:
            b = [b["pixels"][i][b["mask"][i]] for i in range(nchannels)]
        else:
            b = b["pixels"]

        out = np.empty(shape=a.shape)
        for i in range(nchannels):
            if b[i].size == 0:
                out[i] = a[i]
            else:
                out[i, 0] = min(a[i, 0], np.min(b[i]))
                out[i, 1] = max(a[i, 1], np.max(b[i]))
        return out

    def final_minmax(a, b):
        out = np.empty(shape=a.shape)
        for i in range(nchannels):
            out[i, 0] = min(a[i, 0], b[i, 0])
            out[i, 1] = max(a[i, 1], b[i, 1])
        return out

    init = np.empty(shape=(nchannels, 2))
    init[:, 0] = np.inf
    init[:, 1] = -np.inf
    out = bag.foldby(
        key="group",
        binop=combine_extent_partition,
        combine=final_minmax,
        initial=init,
        combine_initial=init
    )

    return out


@check
def sample_normalization(sample, quantiles):
    """
    Perform min-max normalization using quantiles on original pixel data,
    masked pixel data and flat masked intensities list

    Args:
        sample (dict): dictionary containing image data and mask data
        qq: (lower, upper) list of quantiles for every channel
    Returns:
        dict: dictionary including normalized data
    """

    qq = dict(quantiles)[sample["group"]]

    pixels = np.empty_like(sample["pixels"])
    for i in range(len(sample["pixels"])):
        pixels[i] = (sample["pixels"][i] - qq[i, 0]) / (qq[i, 1] - qq[i, 0])

    newsample = copy_without(sample, without="pixels")
    newsample["pixels"] = pixels

    return newsample


def quantile_normalization(images: dask.bag.Bag, nchannels):
    """
    Apply min-max normalization on all images, both on original pixel data and masked pixel data

    Args:
        images (dask.bag): bag of dictionaries containing image data
    Returns:
        dask.bag: bag of dictionaries including normalized data
    """

    def normalize_partition(part, quantiles):
        return [sample_normalization(p, quantiles) for p in part]

    quantiles = get_distributed_minmax(images, nchannels)
    images = images.map_partitions(
        normalize_partition, quantiles.to_delayed()[0])

    return images, quantiles



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterable, Mapping, Optional

import numpy
import dask
import dask.bag
import dask.dataframe

from .shape import shape_features, _shape_features_meta
from .intensity import intensity_features, _intensity_features_meta
from .texture import texture_features, _texture_features_meta
from .bbox import bbox_features, _bbox_features_meta


def compute_features(images, channel_names, types, loader_meta, prefix):

    def rename(c):
        if any(c.startswith(a) for a in list(loader_meta.keys())):
            return f"meta_{c}"
        elif any(c.startswith(a) for a in ["bbox", "regions"]):
            if prefix is not None:
                return f"meta_{prefix}_{c}"
            else:
                return f"meta_{c}"
        else:
            if prefix is not None:
                return f"feat_{prefix}_{c}"
            else:
                return f"feat_{c}"

    features = extract_features(
        images=images,
        channel_names=channel_names,
        types=types,
        loader_meta=loader_meta
    )
    features = features.rename(columns=rename)
    return features


def features_partition(
    part: Iterable[dict],
    *,
    loader_meta_keys: Iterable[str],
    types: Iterable[str],
    lengths: Mapping[str, int]
):

    out = numpy.full(shape=(len(part), lengths["full"]), fill_value=None, dtype=object)
    for i, p in enumerate(part):
        for j, k in enumerate(loader_meta_keys):
            out[i, j] = p[k]
        c = len(loader_meta_keys)

        if "pixels" in p:
            if "bbox" in types:
                out[i, c:c + lengths["bbox"]] = bbox_features(p)
                c += lengths["bbox"]
            if "shape" in types:
                out[i, c:c + lengths["shape"]] = shape_features(
                    mask=p["mask"],
                    combined_mask=p["combined_mask"]
                )
                c += lengths["shape"]
            if "intensity" in types:
                out[i, c:c + lengths["intensity"]] = intensity_features(
                    pixels=p["pixels"],
                    mask=p["mask"],
                    combined_mask=p["combined_mask"],
                    background=p["background"],
                    combined_background=p["combined_background"]
                )
                c += lengths["intensity"]
            if "texture" in types:
                out[i, c:c + lengths["texture"]] = texture_features(p)

    return out


def extract_features(  # noqa: C901
    *,
    images: dask.bag.Bag,
    channel_names: list,
    types: list,
    loader_meta: Optional[Mapping[str, type]] = {}
) -> dask.dataframe.DataFrame:
    """Extracts requested features from pixel values in images.

    Keyword Args:
        images: bag of mappings containing image data. Check each feature
          extraction method (:func:`bbox_features`,
          :func:`scip.features.intensity.intensity_features`,
          :func:`scip.features.shape.shape_features` and
          :func:`scip.features.texture.texture_features`)
          to see what keys must be present in each mapping.
        channel_names: names of channels in the image.
        types: feature types to be extracted from the image.
        maximum_pixel_value: theoretical maximal value in the image.
        loader_meta: data type mapping of meta keys extracted
          by the loader. Defaults to {}.

    Returns:
        dataframe containing all extracted features (columns) for all
          images (rows) in the input bag.
    """

    metas = {}
    if "bbox" in types:
        metas["bbox"] = _bbox_features_meta(channel_names)
    if "shape" in types:
        metas["shape"] = _shape_features_meta(channel_names)
    if "intensity" in types:
        metas["intensity"] = _intensity_features_meta(channel_names)
    if "texture" in types:
        metas["texture"] = _texture_features_meta(channel_names)

    full_meta = loader_meta.copy()
    lengths = {}
    for k, v in metas.items():
        full_meta.update(v)
        lengths[k] = len(v)
    lengths["full"] = len(full_meta)

    images = images.map_partitions(
        features_partition,
        loader_meta_keys=list(loader_meta.keys()),
        types=types,
        lengths=lengths
    )

    images_df = images.to_dataframe(meta=full_meta, optimize_graph=False)

    return images_df



import pytest
from scip import main
from pathlib import Path
import pyarrow.parquet
import pandas
import numpy


@pytest.mark.parametrize(
    'mode,limit,expected_n,with_replacement',
    [
        pytest.param('local', 2, 2, False, marks=pytest.mark.mpi_skip),
        pytest.param('local', 8, 8, False, marks=pytest.mark.mpi_skip),
        pytest.param('local', -1, 10, False, marks=pytest.mark.mpi_skip),
        pytest.param('local', 20, 20, True, marks=pytest.mark.mpi_skip),
        pytest.param('mpi', -1, 10, False, marks=[
            pytest.mark.mpi, pytest.mark.skip(reason="Issues with MPI")])
    ])
def test_main(mode, limit, expected_n, with_replacement, zarr_path, tmp_path, data):
    runtime = main.main(
        mode=mode,
        n_workers=4,
        n_threads=1,
        headless=True,
        output=Path(tmp_path),
        paths=[str(zarr_path)],
        config=data / "scip_zarr.yml",
        limit=limit,
        n_partitions=2,
        with_replacement=with_replacement
    )

    assert runtime is not None
    assert len([f for f in tmp_path.glob("*.parquet")]) == 10
    assert (tmp_path / "scip.log").exists()

    df = pandas.concat(
        [pyarrow.parquet.read_table(f).to_pandas() for f in tmp_path.glob("*.parquet")], axis=0)
    assert len(df) == expected_n

    cols = df.columns
    assert sum("circle-1" in a for a in cols) > 0
    assert sum("circle-2" in a for a in cols) > 0
    assert sum("spot" in a for a in cols) > 0
    assert sum("circle-1" in a for a in cols) == sum("circle-2" in a for a in cols)
    assert sum("circle-1" in a for a in cols) == sum("spot" in a for a in cols)

    assert numpy.all(df.filter(regex="circle-1").values == df.filter(regex="circle-2").values)


def test_main_with_correction(tiffs_folder, tmp_path, data):
    runtime = main.main(
        mode="local",
        n_workers=4,
        n_threads=1,
        headless=True,
        output=Path(tmp_path),
        paths=[str(tiffs_folder)],
        config=data / "scip_tiff_seg.yml",
        n_partitions=2
    )

    assert runtime is not None
    assert len([f for f in tmp_path.glob("*.parquet")]) == 10
    assert (tmp_path / "scip.log").exists()

    df = pandas.concat(
        [pyarrow.parquet.read_table(f).to_pandas() for f in tmp_path.glob("*.parquet")], axis=0)
    assert len(df) > 0

    assert (tmp_path / "correction_images.pickle").exists()



from click.testing import CliRunner
from scip.main import cli
import json
import pandas
import pyarrow.parquet


def test_cli(zarr_path, tmp_path, data):
    runner = CliRunner()
    result = runner.invoke(cli, [
        "--mode", "local",
        "--headless",
        "--n-workers", "4",
        "--n-threads", "1",
        "--n-partitions", "5",
        "--timing", str(tmp_path / "timing.json"),
        str(tmp_path),
        str(data / "scip_zarr.yml"),
        str(zarr_path)
    ])

    assert result.exit_code == 0
    assert (tmp_path / "timing.json").exists()
    with open(tmp_path / "timing.json", "r") as fh:
        timing = json.load(fh)
    assert type(timing["runtime"]) is float
    assert timing["runtime"] > 0

    assert len([f for f in tmp_path.glob("*.parquet")]) == 10
    assert (tmp_path / "scip.log").exists()

    df = pandas.concat(
        [pyarrow.parquet.read_table(f).to_pandas() for f in tmp_path.glob("*.parquet")], axis=0)
    assert len(df) == 10



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Mapping, List

import numpy
from skimage.measure import label, regionprops_table


prop_names = [
    "area",
    "convex_area",
    "eccentricity",
    "equivalent_diameter",
    "euler_number",
    "feret_diameter_max",
    "filled_area",
    "major_axis_length",
    "minor_axis_length",
    "orientation",
    "perimeter",
    "perimeter_crofton",
    "solidity",
    "extent",
    "inertia_tensor-0-0",
    "inertia_tensor-0-1",
    "inertia_tensor-1-0",
    "inertia_tensor-1-1",
    "inertia_tensor_eigvals-0",
    "inertia_tensor_eigvals-1",
    "moments-0-0",
    "moments-0-1",
    "moments-0-2",
    "moments-0-3",
    "moments-1-0",
    "moments-1-1",
    "moments-1-2",
    "moments-1-3",
    "moments-2-0",
    "moments-2-1",
    "moments-2-2",
    "moments-2-3",
    "moments-3-0",
    "moments-3-1",
    "moments-3-2",
    "moments-3-3",
    "moments_central-0-0",
    "moments_central-0-1",
    "moments_central-0-2",
    "moments_central-0-3",
    "moments_central-1-0",
    "moments_central-1-1",
    "moments_central-1-2",
    "moments_central-1-3",
    "moments_central-2-0",
    "moments_central-2-1",
    "moments_central-2-2",
    "moments_central-2-3",
    "moments_central-3-0",
    "moments_central-3-1",
    "moments_central-3-2",
    "moments_central-3-3",
    "moments_hu-0",
    "moments_hu-1",
    "moments_hu-2",
    "moments_hu-3",
    "moments_hu-4",
    "moments_hu-5",
    "moments_hu-6"
]
prop_ids = [
    "area",
    "convex_area",
    "eccentricity",
    "equivalent_diameter",
    "euler_number",
    "feret_diameter_max",
    "filled_area",
    "major_axis_length",
    "minor_axis_length",
    "orientation",
    "perimeter",
    "perimeter_crofton",
    "solidity",
    "extent",
    "inertia_tensor",
    "inertia_tensor_eigvals",
    "moments",
    "moments_central",
    "moments_hu"
]


def _shape_features_meta(channel_names: List[str]) -> Mapping[str, type]:
    out = {}
    for name in ["combined"] + channel_names:
        out.update({f"{p}_{name}": float for p in prop_names})
    return out


def _row(mask: numpy.ndarray) -> numpy.ndarray:
    label_img = label(mask)
    props = regionprops_table(label_image=label_img, properties=prop_ids, cache=False)
    return [numpy.mean(props[k]) for k in prop_names]


def shape_features(
    mask: numpy.ndarray,
    combined_mask: numpy.ndarray
) -> numpy.ndarray:
    """Extracts shape features from image.

    The shape features are extracted using :func:regionprops from scikit-image. These include
    features like eccentricity, convex area or equivalent diameter.

    Args:
        sample (Mapping[str, Any]): mapping with mask and combined mask keys.

    Returns:
        Mapping[str, Any]: extracted shape features.
    """

    out = numpy.full(shape=(len(prop_names * (len(mask) + 1)),), fill_value=None, dtype=float)
    out[:len(prop_names)] = _row(combined_mask)

    for i in range(len(mask)):
        if numpy.any(mask[i]):
            out[(i + 1) * len(prop_names):(i + 2) * len(prop_names)] = _row(mask[i])
        else:
            # setting proper default values if possible when the mask is empty
            out[(i + 1) * len(prop_names):(i + 2) * len(prop_names)] = [
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None
            ]

    return out



"""
Illumination correction as proposed by Jones et al. (2006)
"""

from functools import partial
import numpy
import dask.bag
import dask.delayed
from scipy.signal import medfilt2d
from scipy.ndimage import median_filter
from pathlib import Path
import pickle
import dask.graph_manipulation
from skimage.transform import downscale_local_mean, rescale
import copy


def _binop(total, x):
    if total["pixels"] is None:
        total["pixels"] = numpy.zeros_like(x["pixels"])

    return dict(
        pixels=total["pixels"] + x["pixels"],
        count=total["count"] + 1
    )


def _combine(total1, total2):
    if total1["pixels"] is None:
        total1["pixels"] = numpy.zeros_like(total2["pixels"])

    return dict(
        pixels=total2["pixels"] + total1["pixels"],
        count=total1["count"] + total2["count"]
    )


def _finish(total, filter_func, downscale):
    avg = total[1]["pixels"] / total[1]["count"]

    if downscale > 1:
        # downscale the image prior to median filtering to reduce memory consumption
        avg = downscale_local_mean(avg, factors=(1, downscale, downscale))

    avg = numpy.asarray([
        filter_func(avg[i])
        for i in range(len(total[1]["pixels"]))
    ])
    avg = numpy.where(avg == 0, 1, avg)  # swap out 0 for division no-op 1

    if downscale > 1:
        # reverse earlier downscaling
        avg = rescale(avg, scale=downscale, anti_aliasing=True, channel_axis=0)

    return (total[0], avg)


def correct(
    *,
    images: dask.bag.Bag,
    key: str,
    ngroups: int,
    median_filter_size: int = 50,
    downscale: int = 1,
    output: Path = None,
    precomputed: Path = None
) -> dask.bag.Bag:
    """
    Distributed implementation of Jones et al. (2006) illumination correction. All images
    are averaged per batch, after which the image is filtered using a median filter. If requested,
    the image is downscaled prior to median filtering to reduce memory consumption.

    Args:
        images: Collection containing images to be corrected. Each item in the collection a
            pixels and 'key' key.
        key: Item key used for grouping.
        ngroups: Number of groups in the images collection.
        median_filter_size: Size of the window used in the median filter.
        downscale: factor by which to downscale the image prior to median filtering
        output: Path pointing to directory to save correction images.
        precomputed: Path to pickle file with precomputed correction images in a dict.

    Returns:
        Collection with corrected images.
    """

    # switch to medfilt2d for larger filter sizes as it consumes less memory
    if median_filter_size > 150:
        filter_func = partial(medfilt2d, kernel_size=median_filter_size)
    else:
        filter_func = partial(median_filter, size=median_filter_size)

    def divide(part, mu):
        newpart = copy.deepcopy(part)
        for x in newpart:
            x["pixels"] = x["pixels"] / mu[x[key]]
        return newpart

    if precomputed is not None:
        @dask.delayed(pure=True)
        def load_images(p):
            with open(p, "rb") as fh:
                return pickle.load(fh)
        mean_images = load_images(str(precomputed))
    else:
        mean_images = images.foldby(
            key=key,
            binop=_binop,
            combine=_combine,
            initial=dict(pixels=None, count=0),
            combine_initial=dict(pixels=None, count=0)
        )
        mean_images = mean_images.repartition(npartitions=ngroups)
        mean_images = mean_images.map(_finish, filter_func=filter_func, downscale=downscale)
        mean_images = dask.delayed(dict, pure=True)(mean_images)

    images = images.map_partitions(divide, mu=mean_images)

    if output is not None:
        @dask.delayed
        def save(mean_images):
            with open(str(output / "correction_images.pickle"), "wb") as fh:
                pickle.dump(mean_images, fh)
        return dask.graph_manipulation.bind(
            children=images, parents=save(mean_images), omit=mean_images)

    return images



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Mapping, List

import numpy
import scipy.stats
from skimage.morphology import disk, binary_erosion
from numba import njit

props = [
    'mean',
    'median',
    'max',
    'min',
    'std',
    'mad',
    'lower_quartile',
    'upper_quartile',
    'sum',
    'skewness',
    'kurtosis',
]


def _intensity_features_meta(channel_names: List[str]) -> Mapping[str, type]:
    out = {}
    for i in channel_names:
        out.update({f"{p}_{i}": float for p in props})
        out.update({f"bgcorr_{p}_{i}": float for p in props})
        out.update({f"edge_{p}_{i}": float for p in props})
        out.update({f"bgcorr_edge_{p}_{i}": float for p in props})
        out.update({f"combined_{p}_{i}": float for p in props})
        out.update({f"combined_bgcorr_{p}_{i}": float for p in props})
        out.update({f"combined_edge_{p}_{i}": float for p in props})
        out.update({f"combined_bgcorr_edge_{p}_{i}": float for p in props})
    return out


@njit(cache=True)
def _row(pixels: numpy.ndarray) -> list:
    percentiles = numpy.percentile(pixels, q=(0, 25, 50, 75, 100))

    d = [
        numpy.mean(pixels),
        percentiles[2],
        percentiles[4],
        percentiles[0],
        numpy.std(pixels),
        numpy.median(numpy.absolute(pixels - percentiles[2])),
        percentiles[1],
        percentiles[3],
        numpy.sum(pixels)
    ]

    return d


def _row2(pixels: numpy.ndarray) -> list:
    return [
        scipy.stats.skew(pixels),
        scipy.stats.kurtosis(pixels)
    ]


def _get_edge_mask(mask: numpy.ndarray) -> numpy.ndarray:
    return numpy.bitwise_xor(binary_erosion(mask, footprint=disk(6)), mask)


def intensity_features(
    pixels: numpy.ndarray,
    mask: numpy.ndarray,
    combined_mask: numpy.ndarray,
    background: numpy.ndarray,
    combined_background: numpy.ndarray,
) -> numpy.ndarray:
    """Compute intensity features.

    Find following intensity features based on masked pixel values:
        * 'mean',
        * 'max',
        * 'min',
        * 'var',
        * 'mad',
        * 'skewness',
        * 'kurtosis',
        * 'sum',
        * 'modulation'

    The features are computed on 8 different views on the pixel data:
        1. Raw values of channel specific mask
        2. Background substracted values of channel specific mask
        3. Edge values of channel specific mask
        4. Background substracted edge values of channel specific mask
        5. Raw values of union of masks
        6. Background substracted values of union of masks
        7. Edge values of union of masks
        8. Background substracted edge values of union of masks

    Args:
        sample (Mapping): mapping with pixels, mask, combined_mask, background and
          combined background keys.

    Returns:
        Mapping[str, Any]: extracted features
    """

    out = numpy.full(shape=(len(pixels), 8, len(props)), fill_value=None, dtype=float)

    for i in range(len(pixels)):

        # compute features on channel specific mask
        if numpy.any(mask[i]):

            mask_pixels = pixels[i][mask[i]]
            mask_bgcorr_pixels = mask_pixels - background[i]

            out[i, 0] = _row(mask_pixels) + _row2(mask_pixels)
            out[i, 1] = _row(mask_bgcorr_pixels) + _row2(mask_bgcorr_pixels)

            edge = _get_edge_mask(mask[i])
            if edge.any():
                mask_edge_pixels = pixels[i][edge]
                mask_bgcorr_edge_pixels = mask_edge_pixels - background[i]
                out[i, 2] = _row(mask_edge_pixels) + _row2(mask_edge_pixels)
                out[i, 3] = _row(mask_bgcorr_edge_pixels) + _row2(mask_bgcorr_edge_pixels)
            else:
                out[i, 2:] = 0
        else:
            # write default values
            out[i, :4] = 0

        # always compute features on combined mask (it can never be empty)
        mask_pixels = pixels[i][combined_mask]
        mask_bgcorr_pixels = mask_pixels - combined_background[i]

        out[i, 4] = _row(mask_pixels) + _row2(mask_pixels)
        out[i, 5] = _row(mask_bgcorr_pixels) + _row2(mask_bgcorr_pixels)

        combined_edge = _get_edge_mask(combined_mask)
        if combined_edge.any():
            mask_edge_pixels = pixels[i][combined_edge]
            mask_bgcorr_edge_pixels = mask_edge_pixels - combined_background[i]

            out[i, 6] = _row(mask_edge_pixels) + _row2(mask_edge_pixels)
            out[i, 7] = _row(mask_bgcorr_edge_pixels) + _row2(mask_bgcorr_edge_pixels)
        else:
            out[i, 6:] = 0

    return out.flatten()



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Mapping, List

import numpy
from skimage.feature import graycomatrix, graycoprops
from skimage.filters import sobel


distances = [3, 5]
graycoprop_names = ['contrast', 'dissimilarity', 'homogeneity', 'energy', 'correlation', 'ASM']


def _texture_features_meta(channel_names: List[str]) -> Mapping[str, Any]:

    out = {}
    for i in channel_names:
        for p in graycoprop_names:
            out.update({f"glcm_mean_{p}_{n}_{i}": float for n in distances})
            out.update({f"glcm_std_{p}_{n}_{i}": float for n in distances})
        out[f"sobel_mean_{i}"] = float
        out[f"sobel_std_{i}"] = float
        out[f"sobel_max_{i}"] = float
        out[f"sobel_min_{i}"] = float
        for p in graycoprop_names:
            out.update({f"combined_glcm_mean_{p}_{n}_{i}": float for n in distances})
            out.update({f"combined_glcm_std_{p}_{n}_{i}": float for n in distances})
        out[f"combined_sobel_mean_{i}"] = float
        out[f"combined_sobel_std_{i}"] = float
        out[f"combined_sobel_max_{i}"] = float
        out[f"combined_sobel_min_{i}"] = float
    return out


def _row(pixels, num_features):
    bins = 15
    angles = [
        numpy.pi / 4,  # 45 degrees
        3 * numpy.pi / 4,  # 135 degrees
        5 * numpy.pi / 4,  # 225 degrees
        7 * numpy.pi / 4  # 315 degrees
    ]

    r = (numpy.nanmin(pixels), numpy.nanmax(pixels))
    bin_edges = numpy.histogram_bin_edges(pixels, bins=bins, range=r)
    int_img = numpy.digitize(pixels, bins=bin_edges, right=True)
    glcm = graycomatrix(
        int_img,
        distances=distances,
        angles=angles,
        levels=bins + 2,
        symmetric=True
    )[:-1, :-1]

    out = numpy.full(shape=(num_features,), fill_value=None, dtype=float)
    step = len(distances) * 2
    for i, prop in enumerate(graycoprop_names):
        v = graycoprops(glcm, prop=prop)
        out[i * step:i * step + len(distances)] = v.mean(axis=1)
        out[i * step + len(distances):(i + 1) * step] = v.std(axis=1)

    s = sobel(pixels)
    if numpy.all(numpy.isnan(s)):
        # eventhough at this stage the mask is never empty, the sobel map could be
        # all NaN. For really small cells the kernel will always contain at least
        # one NaN value

        out[-4] = numpy.nan
        out[-3] = numpy.nan
        out[-2] = numpy.nan
        out[-1] = numpy.nan
    else:
        out[-4] = numpy.nanmean(s)
        out[-3] = numpy.nanstd(s)
        out[-2] = numpy.nanmax(s)
        out[-1] = numpy.nanmin(s)

    return out


def texture_features(
    sample: Mapping[str, Any]
):
    """Extracts texture features from image.

    Texture features are computed based on the gray co-occurence level matrix and sobel map.
    From the former, a contrast, dissimilarity, homogeneity, energy, correlation and ASM metric
    is computed. From the latter, mean, standard deviation, maximum and minimum values are computed.

    Features are not computed on background-substracted values (as in
    :func:scip.features.intensity.intensity_features), because all features are computed on
    relative changes of neighbouring pixels; a substraction does not influence these values.

    Args:
        sample (Mapping[str, Any]): mapping with pixels, mask and combined mask keys.

    Returns:
        Mapping[str, Any]: extacted features.

    """
    num_features = len(graycoprop_names) * 2 * len(distances) + 4

    out = numpy.full(shape=(len(sample["pixels"]), 2, num_features), fill_value=None, dtype=float)

    mask_pixels = numpy.where(sample["mask"], sample["pixels"], numpy.nan)
    combined_mask_pixels = numpy.where(sample["combined_mask"], sample["pixels"], numpy.nan)
    for i in range(len(sample["pixels"])):

        # compute features on channel specific mask
        if numpy.any(sample["mask"][i]):
            out[i, 0] = _row(mask_pixels[i], num_features)

        # always compute features on combined mask (it can never be empty)
        out[i, 1] = _row(combined_mask_pixels[i], num_features)

    return out.flatten()



from typing import Mapping

from scip.utils.util import copy_without
import dask.bag
import scipy.stats


def feature_partition(part):
    for i in range(len(part)):
        if "pixels" in part[i]:
            part[i]["filter_sum"] = part[i]["pixels"][0].sum()
    return part


def item(bag: dask.bag.Bag) -> Mapping[str, dask.bag.Item]:
    bag = bag.filter(lambda a: "filter_sum" in a).map(lambda a: a["filter_sum"])
    mu = bag.mean()
    std = bag.std()
    return dict(mu=mu, std=std)


def predicate(x, *, mu, std):
    q5 = scipy.stats.norm.ppf(0.05, loc=mu, scale=std)
    if ("filter_sum" in x) and (x["filter_sum"] > q5):
        return x
    else:
        return copy_without(x, without=["mask", "pixels"])



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

"""
Data loader for zarr format. Based on zarr package. Loader expects each element in the zarr
collection to be a 1-D array, with a corresponding entry in the shape attribute of the zarr
collection. See Input section of documentation for more info.
"""

from typing import Mapping, List

import dask
import dask.bag
import dask.dataframe
import zarr
import numpy
import re
import copy


def reload_image_partition(
    partition: List,
    channels: List[int],
    regex: str
):
    z = zarr.open(partition[0]["path"])
    indices = [p["zarr_idx"] for p in partition]
    data = z.get_coordinate_selection(indices)
    shapes = numpy.array(z.attrs["shape"])[indices]

    newpartition = copy.deepcopy(partition)
    for i in range(len(partition)):
        if "mask" in partition[i]:
            newpartition[i]["pixels"] = data[i].reshape(shapes[i])[channels].astype(numpy.float32)
    return newpartition


def load_image_partition(partition, channels):

    newpartition = copy.deepcopy(partition)
    for i, part in enumerate(partition):
        z = zarr.open(part["path"])
        tmp = z[part["zarr_idx"]]
        tmp = tmp.reshape(z.attrs["shape"][part["zarr_idx"]])[channels]
        newpartition[i]["pixels"] = tmp

    return newpartition

    # start, end = partition[0]["zarr_idx"], partition[-1]["zarr_idx"]
    # data = z[start:end + 1]
    # shapes = z.attrs["shape"][start:end + 1]

    # for i in range(len(partition)):
    #     partition[i]["pixels"] = data[i].reshape(shapes[i])[channels].astype(numpy.float32)

    # return partition


def get_loader_meta(
    *,
    regex: str,
    **kwargs
) -> Mapping[str, type]:
    loader_meta = dict(path=str, zarr_idx=int, object_number=int)
    named_groups = re.findall(r"\(\?P\<([^>]+)\>[^)]+\)", regex)
    for k in named_groups:
        loader_meta[k] = str
    return loader_meta


def get_group_keys():
    return []


@dask.delayed
def meta_from_directory(path, regex):
    match = re.search(regex, str(path))
    groups = match.groupdict()

    z = zarr.open(path, mode="r")

    events = []
    for i, obj in enumerate(z.attrs["object_number"]):
        events.append({**groups, **{
            "path": path,
            "zarr_idx": i,
            "object_number": obj
        }})

    return events


def load_pixels(
    images: dask.bag.Bag,
    channels: List[int],
    **kwargs
) -> dask.bag.Bag:

    if channels is None:
        channels = numpy.s_[:]
    return images.map_partitions(load_image_partition, channels)



import logging
from typing import List, Tuple, Mapping, Any

import dask.bag

from pathlib import Path
from functools import partial


def _load_image_partition(partition, channels, load):
    return [load(event, channels) for event in partition]


def get_images_bag(
    *,
    paths: List[str],
    output: Path,
    channels: List[int],
    config: Mapping[str, Any],
    partition_size: int,
    gpu_accelerated: bool,
    loader_module
) -> Tuple[dask.bag.Bag, Mapping[str, type]]:

    loader = partial(
        loader_module.bag_from_directory,
        channels=channels,
        partition_size=partition_size,
        gpu_accelerated=gpu_accelerated,
        output=output,
        **(config["load"]["kwargs"] or dict())
    )

    images = []

    loader_meta = loader_module.get_loader_meta(**(config["load"]["kwargs"] or dict()))

    for path in paths:
        assert Path(path).exists(), f"{path} does not exist."

        logging.info(f"Bagging {path}")
        bag = loader(path=path)
        images.append(bag)

    images = dask.bag.concat(images)
    return images, loader_meta



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

"""
Data loader for Carl Zeiss Image format. Based on the aicsimageio package. Expects scenes and CZXY
dimensions to be present.
"""

import re
from typing import List, Mapping, Any

from aicsimageio import AICSImage
import dask.bag
import dask.dataframe
import dask.array
import dask

from scip.loading import util as l_util


def _load_block(event, channels):
    im = AICSImage(event["path"], reconstruct_mosaic=False)
    im.set_scene(event["scene"])

    newevent = event.copy()

    if channels is not None:
        newevent["pixels"] = im.get_image_data("CZXY", T=0, C=channels)
    else:
        newevent["pixels"] = im.get_image_data("CZXY", T=0)

    return newevent


def get_loader_meta(**kwargs) -> Mapping[str, type]:
    return dict(path=str, tile=int, scene=str, id=int)


def get_group_keys():
    return ["scene", "tile"]


@dask.delayed
def meta_from_directory(
    path: str,
    scenes: List[str]
) -> List[Mapping[str, Any]]:

    im = AICSImage(path, reconstruct_mosaic=False)

    if (scenes is None) or (type(scenes) is str):
        im_scenes = im.scenes
        if type(scenes) is str:
            im_scenes = filter(lambda s: re.match(scenes, s), im_scenes)
    elif type(scenes) is list:
        im_scenes = scenes
    else:
        raise ValueError("Scenes configuration cannot be recognized.")

    scenes_meta = []
    for scene in im_scenes:
        # store the scene and tile name
        scenes_meta.extend([dict(scene=scene, tile=i, path=path) for i in range(im.shape[0])])

    return scenes_meta


def load_pixels(
    images: dask.bag.Bag,
    channels: List[int],
    **kwargs
) -> dask.bag.Bag:
    return images.map_partitions(
        l_util._load_image_partition, channels=channels, load=_load_block)



from typing import Mapping

import dask.dataframe
from dask.highlevelgraph import HighLevelGraph
from pathlib import Path
from concurrent.futures import Future
import pandas
import anndata
import numpy


def export(
    *,
    df: dask.dataframe.DataFrame,
    output: Path,
    filename: str
) -> Future:
    """Exports dataframe to one AnnData .h5ad-file per partition.

    Keyword args:
        df: Dataframe to be exported.
        output: Path to directory where objects should be stored.
        filename: Filename to give to partition
          objects (will be named with format string {filename}.{partition}.h5ad).

    Returns:
        Future that represents the export task.
    """

    def _write_anndata(
        df: pandas.DataFrame,
        partition_info: Mapping[str, int]
    ):
        """Creates and writes AnnData object for one partition"""
        x = partition_info["number"]
        df = df.reset_index(drop=True)
        df.index = df.index.astype(str)
        anndata.AnnData(
            X=df.filter(regex="feat"),
            dtype=numpy.float32,
            obs=df.filter(regex="meta")
        ).write(output / f"{filename}.{x}.h5ad")

    data_write = df.map_partitions(
        _write_anndata,
        meta=df._meta,
        enforce_metadata=False,
        transform_divisions=False,
        align_dataframes=False
    )

    # Add noop to graph
    final_name = "store-" + data_write._name
    dsk = {(final_name, 0): (lambda x: None, data_write.__dask_keys__())}

    # Convert data_write + dsk to computable collection
    graph = HighLevelGraph.from_collections(final_name, dsk, dependencies=(data_write,))
    out = dask.dataframe.core.Scalar(graph, final_name, "")

    return out



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

from skimage.filters import threshold_otsu
from skimage.segmentation import watershed, expand_labels
from skimage.restoration import denoise_nl_means
from skimage import feature, measure
import skimage
from scipy.ndimage import distance_transform_edt
from skimage.measure import regionprops

from typing import List
import dask
import numpy


@dask.delayed
def segment_block(
    block: numpy.ndarray,
    *,
    group: str,
    cell_diameter: int,
    dapi_channel: int
) -> List[dict]:

    plane = block[0, dapi_channel]
    plane = skimage.img_as_float32(plane)

    plane = denoise_nl_means(plane, patch_size=3, patch_distance=2, multichannel=False)

    t = threshold_otsu(plane)
    cells = plane > t
    distance = distance_transform_edt(cells)

    local_max_coords = feature.peak_local_max(distance, min_distance=cell_diameter)
    local_max_mask = numpy.zeros(distance.shape, dtype=bool)
    local_max_mask[tuple(local_max_coords.T)] = True
    markers = measure.label(local_max_mask)

    segmented_cells = watershed(-distance, markers, mask=cells)
    segmented_cells = expand_labels(segmented_cells, distance=cell_diameter * 0.25)

    events = []
    props = regionprops(segmented_cells)
    for prop in props:
        bbox = prop.bbox
        events.append(dict(
            pixels=block[0, :, bbox[0]: bbox[2], bbox[1]:bbox[3]],
            mask=numpy.repeat(prop.image[numpy.newaxis] > 0, block.shape[1], axis=0),
            combined_mask=prop.image > 0,
            group=group,
            bbox=tuple(bbox),
            regions=[1] * block.shape[1],
            background=numpy.zeros(shape=(block.shape[1],), dtype=float),
            combined_background=numpy.zeros(shape=(block.shape[1],), dtype=float)
        ))

    return events



import setuptools
import versioneer

setuptools.setup(
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Image Processing"
    ],
    name="scip",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        'numpy',
        'Pillow',
        'dask>=2022.01.1',
        'dask-ml',
        'click',
        'scikit-image',
        'pyyaml',
        'graphviz',
        'pyarrow',
        'umap-learn',
        'bokeh',
        'zarr',
        'fastparquet',
        'anndata'
    ],
    entry_points={
        'console_scripts': [
            'scip=scip.main:cli'
        ]
    },
    extras_require={
        "mpi": ['dask_mpi', 'mpi4py'],
        "czi": ['cellpose', 'aicsimageio', 'aicspylibczi'],
        "jobqueue": ["dask-jobqueue"]
    },
    python_requires=">=3.8",
    package_data={
        'scip': ['utils/logging.yml']
    }
)



# Copyright (C) 2022 Maxim Lippeveld
#
# This file is part of SCIP.
#
# SCIP is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SCIP is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SCIP.  If not, see <http://www.gnu.org/licenses/>.

"""
Data loader for Tagged Image File format based on aicsimageio package. Expects one channel per file.

Regex pattern for load_pixels and meta_from_directory functions has to contain
channel and id named group.
"""

from typing import Callable, List, Mapping, Any

import re
import logging
from pathlib import Path
from functools import partial

import pandas
import dask
import dask.bag
import dask.dataframe
import dask.array
import numpy
import tifffile
from aicsimageio.readers.tiff_glob_reader import TiffGlobReader
from scip.loading import util as l_util

logging.getLogger("tifffile").setLevel(logging.ERROR)


def _load_image_tiff(
    event: Mapping[str, Any],
    channels: List[int]
):
    """Adds pixels attribute containing pixel values of this event. Values are loaded from
    one file per channel.

    Keyword args:
        event (Mapping[str, Any]): Contains information for an event. Must contain keys
            corresponding to each channel.
        channels (List[int]): List of channels to load.

    Returns:
        event (Mapping[str, Any]): event with pixel values.
    """

    try:
        paths = [event[str(c)] for c in channels]
        arr = tifffile.imread(paths)
        arr = arr.astype(numpy.float32)

        # tifffile collapses axis with size 1,
        # occurrs when only one path is passed
        if len(arr.shape) < 3:
            arr = arr[numpy.newaxis, ...]

        newevent = event.copy()
        newevent["pixels"] = arr
        return newevent
    except TypeError as e:
        logging.getLogger(__name__).exception(e)
        logging.getLogger(__name__).error(paths)
        raise e


def _load_block(
    event: Mapping[str, Any],
    channels: List[int],
    map_to_index: Callable
):
    """Adds pixels attribute containing pixel values of this event. Values are loaded from one
    file per channel.

    Keyword args:
        event (Mapping[str, Any]): Contains information for an event. Must contain keys
            corresponding to each channel.
        channels (List[int]): List of channels to load.
        map_to_index (Callable): callable that takes file name as input and outputs pandas Series
            indicating channel index.

    Returns:
        event (Mapping[str, Any]): event with pixel values.
    """

    paths = [event[str(c)] for c in channels]

    im = TiffGlobReader(
        glob_in=paths,
        indexer=map_to_index
    ).get_image_data("CXY")

    newevent = event.copy()
    newevent["pixels"] = im.astype(numpy.float32)
    return newevent


def get_loader_meta(**kwargs) -> Mapping[str, type]:
    """Returns key to type mapping of metadata."""
    return dict(path=str)


def get_group_keys():
    return []


def _map_to_index(f, regex, channels):
    idx = int(re.search(regex, str(f)).group("channel"))
    m = {c: i for i, c in enumerate(channels)}
    return pandas.Series(dict(S=0, T=0, C=m[idx], Z=0))


@dask.delayed
def meta_from_directory(path, regex):
    logger = logging.getLogger(__name__)

    path = Path(path)

    matches = []
    i = 0
    for p in path.glob("*.tif*"):
        m = re.search(regex, str(p))
        if m is not None:
            groups = m.groupdict()
            matches.append({
                **groups,
                **dict(path=str(p))
            })
            i += 1

    df = pandas.DataFrame.from_dict(matches)
    df1 = df.pivot(index="id", columns="channel", values="path")
    df = df.set_index("id")
    df2 = df.loc[~df.index.duplicated(keep='first'), df.drop(columns=["path", "channel"]).columns]

    df = pandas.concat([df1, df2], axis=1)

    pre_filter = len(df)
    df = df[~df1.isna().any(axis=1)]  # drop rows with missing files
    dropped = pre_filter - len(df)
    logger.warning("Dropped %d rows because of missing channel files in %s" % (dropped, str(path)))

    df["path"] = df.iloc[:, 0]

    return df.to_dict(orient="records")


def load_pixels(
    images: dask.bag.Bag,
    channels: List[int],
    regex: str,
    **kwargs
) -> dask.bag.Bag:
    _m2i = partial(_map_to_index, channels=channels, regex=regex)
    func = partial(_load_block, map_to_index=_m2i)

    return images.map_partitions(l_util._load_image_partition, channels=channels, load=func)

    # bag = dask.bag.from_delayed(meta)
    # bag = bag.map_partitions(
    #     l_util._load_image_partition, channels=channels, load=_load_image_tiff)



from typing import List, Mapping, Any
from pathlib import Path
from importlib import import_module
import numpy
import dask
import dask.bag


def _substract_mask(event, left_index, right_index, for_channel_index):
    event["mask"][for_channel_index] = event["mask"][left_index] - event["mask"][right_index]
    return event


def segment(
    *,
    images: dask.bag.Bag,
    method: str,
    settings: Mapping[str, Any],
    export: bool,
    output: Path,
    gpu: bool,
    loader_module
) -> dask.bag.Bag:

    mod = import_module('scip.segmentation.%s' % method)

    # this segment operation is annotated with the cellpose resource to let the scheduler
    # know that it should only be executed on a worker that also has the cellpose resource.
    with dask.annotate(resources={"cellpose": 1}):
        images = images.map_partitions(mod.segment_block, gpu_accelerated=gpu > 0, **settings)

    if settings["substract"] is not None:
        images = images.map(
            _substract_mask,
            left_index=settings["substract"]["left_index"],
            right_index=settings["substract"]["right_index"],
            for_channel_index=settings["substract"]["for_index"]
        )

    group_keys = loader_module.get_group_keys()
    if export:
        assert len(group_keys) > 0, "At least one group key is required to export the segmentations"
        assert output is not None, "Output path is required to export the segmentations"
        images = images.map_partitions(
            export_labeled_mask,
            out=output, group_keys=group_keys)

    return images.map_partitions(mod.to_events, group_keys=group_keys, **settings)


def export_labeled_mask(
    events: List[Mapping[str, Any]],
    out: Path,
    group_keys: List[str]
) -> List[Mapping[str, Any]]:
    (out / "masks").mkdir(parents=False, exist_ok=True)

    for event in events:
        f = "%s.npy" % "_".join([str(event[k]) for k in group_keys])
        numpy.save(out / "masks" / f, event["mask"])

    return events



from typing import List, Mapping, Any


def _bbox_features_meta(channel_names: List[str]) -> Mapping[str, type]:
    d = {
        "bbox_minr": float,
        "bbox_minc": float,
        "bbox_maxr": float,
        "bbox_maxc": float
    }
    d.update({f"regions_{i}": float for i in channel_names})
    return d


def bbox_features(p: Mapping) -> Mapping[str, Any]:
    """Extracts bbox features from image.

    The bbox consist of four values: bbox_minr, bbox_minc, bbox_maxr, bbox_maxc.

    Args:
        p: Contains a sequence of 4 numbers under key bbox.

    Returns:
        Extracted features.
    """

    return list(p["bbox"]) + p["regions"]
